#include "vec3.h"

vec3 vec3::random_unit_vector()
{
    auto a = math::random_double(0, 2 * math::pi);
    auto z = math::random_double(-1,1);
    auto r = sqrt(1 - z*z);
    return vec3(r*cos(a), r*sin(a), z);
}

vec3 vec3::reflect(const vec3& v, const vec3& n)
{
    return v - 2*dot(v,n)*n;
}

vec3 vec3::refract(const vec3& uv, const vec3& n, double etai_over_etat)
{
    auto cos_theta = dot(-uv, n);
    vec3 r_out_parallel = etai_over_etat * (uv + cos_theta*n);
    vec3 r_out_perp = -sqrt(1.0 - r_out_parallel.length_squared()) * n;

    return r_out_parallel + r_out_perp;
}

vec3 vec3::random_in_unit_disk()
{
    while(true)
    {
        auto p = vec3(math::random_double(-1,1), math::random_double(-1,1), 0);
        if (p.length_squared() >= 1) continue;
        return p;
    }
}

vec3 vec3::random_in_unit_sphere()
{
    while (true)
    {
        auto p = vec3::random(-1,1);
        if (p.length_squared() >= 1) continue;
        return p;
    }
}

vec3 random_in_hemisphere(const vec3& normal)
{
    vec3 in_unit_sphere = vec3::random_in_unit_sphere();
    if (dot(in_unit_sphere, normal) > 0.0)
    {
        return in_unit_sphere;
    }
    else
    {
        return -in_unit_sphere;
    }
}

